(*********************************************************************

  HOC13Bisimulations

**********************************************************************

  Copyright (C) 2014
  Simon Boulier, Martin Escarra, Petar Maksimovic, Alan Schmitt

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

**********************************************************************)

Require Import HOC01Defs.
Require Import HOC03CanonicalLemmas.
Require Import HOC04DefLTS.
Require Import HOC05WFProcesses.
Require Import HOC06FreshLemmas.
Require Import HOC07SizeLemmas.
Require Import HOC08SubstLemmas.
Require Import HOC09GVLemmas.
Require Import HOC10CongrLemmas.
Require Import HOC11TransLemmas.

Require Import Relations.

(******************************* RELATIONS *******************************)

Definition in_relation (R : RelWfP) : Prop := 
  forall (p q : wfprocess), (R p q) ->
    forall (a : chan) (fp : abstraction), {{p -- LabIn a ->> (AA fp)}} ->
    exists (fq : abstraction), {{q -- LabIn a ->> (AA fq)}} /\
      forall X, X#p -> X#q -> 
        R (wfp_inst_Abs fp (wfp_Gvar X)) (wfp_inst_Abs fq (wfp_Gvar X)).

Definition in_full_relation (R : RelWfP) : Prop := 
  forall (p q : wfprocess), (R p q) ->
    forall (a : chan) (fp : abstraction), {{p -- LabIn a ->> (AA fp)}} -> 
    exists (fq : abstraction), {{q -- LabIn a ->> (AA fq)}} /\
      forall X, R (wfp_inst_Abs fp (wfp_Gvar X)) (wfp_inst_Abs fq (wfp_Gvar X)).

Definition in_normal_relation (R : RelWfP) : Prop :=
  forall p q, (R p q) -> forall a fp, {{p--LabIn a->>(AA fp)}} ->
  exists fq, {{q--LabIn a->>(AA fq)}} /\ forall m : chan, m###p -> m###q -> 
    (R (wfp_inst_Abs fp (wfp_Send m wfp_Nil)) (wfp_inst_Abs fq (wfp_Send m wfp_Nil))).

Definition out_relation (R : RelWfP) : Prop :=
  forall (p q : wfprocess), (R p q) -> 
    forall (a : chan) (p' p'' : wfprocess), 
    {{p -- LabOut a p'' ->> (AP p')}} ->
      exists (q' q'' : wfprocess),
        ({{q -- LabOut a q'' ->> (AP q')}} /\ (R p' q') /\ (R p'' q'')).

Definition out_normal_relation (R : RelWfP) : Prop :=
  forall p q, (R p q) -> 
    forall a (p' : wfprocess) (p'' : wfprocess), {{p--LabOut a p''->>(AP p')}} ->  
      exists (q' : wfprocess), exists (q'' : wfprocess), 
        ({{q--LabOut a q''->>(AP q')}}  /\
        forall (m:chan) (X:var), m###p -> m###q -> X#p'' -> X#q'' ->  
          (R (wfp_Par (wfp_Abs m X p'') p')) (wfp_Par (wfp_Abs m X q'') q')).

Definition out_context_relation (R : RelWfP) : Prop :=
  forall p q, (R p q) -> 
    forall a (p' p'' : wfprocess), {{p--LabOut a p''->>(AP p')}} ->
      exists (q' : wfprocess), exists (q'' : wfprocess), 
        ({{q--LabOut a q''->>(AP q')}}  /\ 
        forall X (s : wfprocess), (GV s) \c (singleton X) ->  
          (R (wfp_Par (wfp_subst p'' X s) p') (wfp_Par (wfp_subst q'' X s) q'))). 

Definition var_relation (R : RelWfP) : Prop :=
  forall (p q : wfprocess), (R p q) -> 
    forall (X : var) (p' : wfprocess), {{p -- LabRem X ->> (AP p')}} -> 
      exists (q' : wfprocess), {{q -- LabRem X ->> (AP q')}} /\ (R p' q').

Definition tau_relation (R : RelWfP) : Prop :=
  forall p q, (R p q) -> 
    forall (p' : wfprocess), {{p--Tau->>(AP p')}} ->  
      exists (q' : wfprocess), ({{q--Tau->>(AP q')}} /\ R p' q').

Definition closed_relation (R : RelWfP) : Prop :=
  forall p q, (R p q) -> 
    forall a fp, {{p--LabIn a->>(AA fp)}} ->
      exists fq, {{q -- LabIn a ->> (AA fq)}} /\
        (forall (r : wfprocess), closed_process r -> 
          R (wfp_inst_Abs fp r) (wfp_inst_Abs fq r)).

(******************************* BISIMULATIONS *******************************)

Definition IO_bisimulation (R : RelWfP) : Prop :=
 (Symmetric R) /\ (in_relation R) /\ (out_relation R) /\ (var_relation R).

Definition IObis (p q : wfprocess) : Prop :=
  exists R, (IO_bisimulation R) /\ (R p q).

Notation "p ≈ q" := (IObis p q) (at level 60).



Definition HigherOrder_bisimulation (R : RelWfP) : Prop :=
  (Symmetric R) /\ (tau_relation R) /\ (out_relation R) /\ (closed_relation R).

Definition HObis (p q: wfprocess) : Prop :=
  exists R, (HigherOrder_bisimulation R) /\ R p q.

Definition HObisC (p q: wfprocess) : Prop :=
  closed_process p /\ closed_process q /\ HObis p q.

Notation "p ≈HO q" := (HObisC p q) (at level 60).



Definition Context_bisimulation (R : RelWfP) : Prop :=
  (Symmetric R) /\ (tau_relation R) /\ (out_context_relation R) /\ (closed_relation R).

Definition CONbis p q : Prop :=
  exists R, (Context_bisimulation R) /\ (R p q).

Definition CONbisC (p q : wfprocess) : Prop :=
  closed_process p /\ closed_process q /\ CONbis p q.

Notation "p ≈CON q" := (CONbisC p q) (at level 60).



Definition Normal_bisimulation (R : RelWfP) : Prop :=
  (Symmetric R) /\ (tau_relation R) /\ (out_normal_relation R) /\ (in_normal_relation R).

Definition NORbis p q : Prop :=
  exists R, (Normal_bisimulation R) /\ (R p q).

Definition NORbisC (p q : wfprocess) : Prop :=
  closed_process p /\ closed_process q /\ NORbis p q.

Notation "p ≈NOR q" := (NORbisC p q) (at level 60).



Definition OpenNormal_bisimulation (R : RelWfP) : Prop :=
  (Symmetric R) /\ (tau_relation R) /\ (out_normal_relation R) /\ (in_relation R) /\ (var_relation R).

Definition ONORbis p q : Prop :=
  exists R, (OpenNormal_bisimulation R) /\ (R p q).

Notation "p ≈ONOR q" := (ONORbis p q) (at level 60).

(*************** CLOSED BISIMILARITIES EXTENSIONS FOR OPEN PROCESSES ***************)

Definition HObisE (p q : wfprocess) : Prop := forall (a:chan) (xs:list var),
  permut xs (to_list (GV p \u GV q)) ->
  (wfp_pipe a xs p) ≈HO (wfp_pipe a xs q).

Notation "p ≈*HO q" := (HObisE p q) (at level 60).



Definition CONbisE (p q : wfprocess) : Prop := forall (a:chan) (xs:list var),
  permut xs (to_list (GV p \u GV q)) ->
  (wfp_pipe a xs p) ≈CON (wfp_pipe a xs q).

Notation "p ≈*CON q" := (CONbisE p q) (at level 60).



Definition NORbisE (p q : wfprocess) : Prop := forall (a:chan) (xs:list var),
  permut xs (to_list (GV p \u GV q)) ->
  (wfp_pipe a xs p) ≈NOR (wfp_pipe a xs q).

Notation "p ≈*NOR q" := (NORbisE p q) (at level 60).

(*************** UP-TO BISIMULATIONS ***************)

(* A relation is fresh if it preserves free variables 
Definition fresh_relation (R:binary wfprocess) : Prop :=
  forall p q, R p q -> forall X, X#p <-> X#q. *)

(* An up-to relation *)
Inductive up_to (Re R: binary wfprocess) : binary wfprocess :=
 | UpTo : forall p q, (exists p' q', Re p p' /\ Re q' q /\ R p' q') -> up_to Re R p q.

Hint Resolve UpTo.

(* If a relation is reflexive, any other relation is automatically up-to it *)
Lemma up_to_eq : forall (Re: binary wfprocess), Reflexive Re ->
  forall (R: binary wfprocess) p q, R p q -> up_to Re R p q.
Proof.
  introv HRe Hr. constructor. exists~ p q.
Qed.

Definition in_relation_up_to (Re R:binary wfprocess) : Prop :=
  forall p q, (R p q) -> forall a fp, {{p--LabIn a->>(AA fp)}} -> 
    exists fq, {{q--LabIn a->>(AA fq)}} /\ 
    forall (X:var), X#p -> X#q -> 
      up_to Re R (wfp_inst_Abs fp (wfp_Gvar X)) (wfp_inst_Abs fq (wfp_Gvar X)).

Definition out_relation_up_to (Re R:binary wfprocess) : Prop :=
  forall p q, (R p q) -> forall a (p' p'' : wfprocess), {{p--LabOut a p''->>(AP p')}} -> exists (q' : wfprocess), exists (q'' : wfprocess), ({{q--LabOut a q''->>(AP q')}} /\ (up_to Re R p' q') /\ (up_to Re R p'' q'')).

Definition var_relation_up_to (Re R:binary wfprocess) : Prop :=
  forall p q, (R p q) -> forall X (p' : wfprocess), {{p--LabRem X->>(AP p')}} ->
exists (q' : wfprocess), {{q--LabRem X->>(AP q')}} /\ (up_to Re R p' q').

Definition tau_relation_up_to (Re R : RelWfP) : Prop :=
  forall p q, (R p q) -> 
    forall (p' : wfprocess), {{p--Tau->>(AP p')}} ->  exists (q' : wfprocess), ({{q--Tau->>(AP q')}} /\ (up_to Re R p' q')).

Definition out_normal_relation_up_to (Re R : RelWfP) : Prop :=
  forall p q, (R p q) -> 
              forall a (p' : wfprocess) (p'' : wfprocess), {{p--LabOut a p''->>(AP p')}} ->  
              exists (q' : wfprocess), exists (q'' : wfprocess), 
              ({{q--LabOut a q''->>(AP q')}}  /\
              forall (m:chan) (X:var), m###p -> m###q -> X#p'' -> X#q'' ->  
              (up_to Re R 
                (wfp_Par (wfp_Abs m X p'') p')) 
                (wfp_Par (wfp_Abs m X q'') q')).

Definition IO_bisimulation_up_to (Re R:binary wfprocess) : Prop :=
  (Symmetric R) /\ (in_relation_up_to Re R) /\ (out_relation_up_to Re R) /\ (var_relation_up_to Re R).

Definition IObis_up_to Re p q : Prop :=
  exists R, (IO_bisimulation_up_to Re R) /\ (R p q).

Definition OpenNormal_bisimulation_up_to (Re R : RelWfP) : Prop :=
  (Symmetric R) /\ (tau_relation_up_to Re R) /\ (out_normal_relation_up_to Re R) /\ 
                   (in_relation_up_to Re R) /\ (var_relation_up_to Re R).

Definition ONORbis_up_to Re p q : Prop :=
  exists R, (OpenNormal_bisimulation_up_to Re R) /\ (R p q).

(**** STRUCTURAL CONGRUENCE ON WFP AND BISIMULATIONS ****)


Lemma congr_in_strong : forall (p q : wfprocess) a fp, 
  p ≡* q -> {{p -- LabIn a ->> AA fp}} -> 
  (exists fq, {{q -- LabIn a ->> AA fq}} /\ 
    forall r : wfprocess, (wfp_inst_Abs fp r) ≡* (wfp_inst_Abs fq r)).
Proof with (introv; simpl; try solve [applys* t_step]; automate).
  introv Hr Ht.

  gen fp; inductions Hr.

  Focus 2. 
  introv Ht.
  lets (fq & Ht' & Hc) : (IHHr1 _ Ht).
    lets (fr & Ht'' & Hc2): (IHHr2  _ Ht'). exists fr. split~.
      introv. etransitivity.
        apply~ Hc. 
        apply~ Hc2. 
 
  inductions H; introv Ht...
  (*send*) inverts Ht. 
  (*receive*) 
  lets Heq : wfp_trans_in_chan_eq Ht; subst.
    lets Hyp : wfp_trans_in_abs_eq Ht; inverts Hyp as Hyp; subst.
    eexists. split. constructor~.
    introv; simpl. applys* wfcgr_subst; constructor~.
    destruct Hyp as (Y & Hneq & Hfy & Heq); subst.
    exists (AbsPure X q). splits~.
      introv; simpl. rewrite~ <- wfp_subst_decomposition_g.
      apply~ wfcgr_subst... 
  (*par*) inverts Ht...
    (*1*) destruct ap; inverts H5. lets (f2 & Ht' & Hc): (IHwfcgr_step1 a0 H4).
    exists (AbsPar1 f2 q'). split...
      introv; simpl. apply wfcgr_par...
    (*2*) destruct aq; inverts H5. lets (f2 & Ht' & Hc): (IHwfcgr_step2 a0 H4).
    exists (AbsPar2 p' f2). split...
      introv; simpl. apply wfcgr_par... 
  (*par nil1*) exists (AbsPar1 fp wfp_Nil). split...
  (*par nil2*) inverts Ht...
      destruct ap; inverts H3. exists a0. split...
      inverts H2.
  (*par com*) inverts Ht...
   (*1*) destruct ap; inverts H3. exists (AbsPar2 q a0). split...
   (*2*) destruct aq; inverts H3. exists (AbsPar1 a0 p). split...
  (*par assoc*) inverts Ht...
    (*1*) destruct ap; inverts H3. exists (AbsPar1 (AbsPar1 a0 q) r). split...
    (*2*) inverts H2...
      destruct ap; inverts H3. exists (AbsPar1 (AbsPar2 p a0) r). split...
    (*2*) destruct aq0; inverts H3. exists (AbsPar2 (wfp_Par p q) a0). split...
   (*par assoc*) inverts Ht...
    (*1*) destruct ap; inverts H3. inverts H2...
      destruct ap; inverts H4. exists (AbsPar1 a1 (wfp_Par q r)). split...
    (*1*) destruct aq; inverts H4. exists (AbsPar2 p (AbsPar1 a1 r)). split...
    (*2*) destruct aq; inverts H3. exists (AbsPar2 p (AbsPar2 q a0)). split...
  (*refl*) exists fp. split...
Qed.

Lemma congr_in : in_relation wfcgr.
Proof.
  introv Hr Ht. 
  lets HDexM : congr_in_strong Hr Ht.
  destruct HDexM as (fq & H' & Heq).
  exists fq; split~. 
Qed.

Lemma congr_out : out_relation wfcgr.
Proof with (try solve [applys* t_step]; automate).
  introv Hr Ht.  
  inductions Hr.

  Focus 2.
  lets (q' & q'' & Ht' & Hc1 & Hc1'): (IHHr1 _ _ _ Ht).
  lets (r' & r'' & Ht'' & Hc2 & Hc2'): (IHHr2 _ _ _ Ht'). 
  exists r' r''. splits~.
    etransitivity; eassumption.
    etransitivity; eassumption. 

  gen p'' p'; inductions H; introv Ht...
  (*send*) inverts Ht... 
    exists wfp_Nil q. splits... 
  (*receive*) inverts Ht. 
  (*par*) inverts Ht...
    (*1*) destruct ap; inverts H5.
    lets (p2' & p2'' & Ht' & Hc1 & Hc2): IHwfcgr_step1 H4.
    exists (wfp_Par p2' q') p2''. splits...
      apply~ wfcgr_par... 
    (*2*) destruct aq; inverts H5.
    lets (q2' & q2'' & Ht' & Hc1 & Hc2): IHwfcgr_step2 H4.
    exists (wfp_Par p' q2') q2''. splits~...
      apply~ wfcgr_par... 
  (*par nil1*) exists (wfp_Par p' wfp_Nil) p''. splits...
  (*par nil2*) inverts Ht... 
    destruct ap; inverts H3. 
      exists w p''; splits...      
      inverts H2.
  (*par com*) inverts Ht...
    (*1*) destruct ap; inverts H3...
      exists (wfp_Par q w) p''. splits...   
    (*2*) destruct aq; inverts H3...
      exists (wfp_Par w p) p''. splits...
  (*par assoc*) inverts Ht... 
    (*1*) destruct ap; inverts H3...
      exists (wfp_Par (wfp_Par w q) r) p''; splits...
    (*2*) inverts H2.
    destruct ap; inverts H3... 
      exists (wfp_Par (wfp_Par p w) r) p''. splits...
    (*2*) destruct aq0; inverts H3... 
      exists (wfp_Par (wfp_Par p q) w) p''. splits...
  (*par assoc*) inverts Ht... 
    (*2*) inverts H2...
    destruct ap0; inverts H3... 
      exists (wfp_Par w (wfp_Par q r)) p''. splits...
    (*2*) destruct aq; inverts H3... 
      exists (wfp_Par p (wfp_Par w r)) p''. splits...
    (*1*) destruct aq; inverts H3...
      exists (wfp_Par p (wfp_Par q w)) p''; splits...
  (*refl*) exists p' p''. splits...
Qed.

Lemma congr_rem : var_relation wfcgr.
Proof with (try solve [applys* t_step]; automate).
  introv Hr Ht; inductions Hr...

  Focus 2.
    lets (r' & Ht' & Hc2): IHHr1 Ht.
    lets (r'' & Ht'' & Hc3): IHHr2 Ht'. 
    exists r''; splits...
      etransitivity; eassumption.

  inductions H; try solve [inverts Ht]...

  (*par*) inverts Ht...
    (*1*) destruct ap; inverts H5.
      lets (p2' & Ht' & Hc): IHwfcgr_step1 H4.
      exists (wfp_Par p2' q'). splits...
        apply~ wfcgr_par... 
    (*2*) destruct aq; inverts H5.
      lets (q1 & Ht' & Hc): IHwfcgr_step2 H4. 
      exists (wfp_Par p' q1). splits...
        apply~ wfcgr_par... 
  (*par nil1*) exists (wfp_Par p' wfp_Nil); splits...
  (*par nil2*) inverts Ht...
    destruct ap; inverts H3. 
    exists w. splits...
    inverts H2.
  (*par com*) inverts Ht...
    (*1*) destruct ap; inverts H3. 
      exists (wfp_Par q w). splits... 
    (*2*) destruct aq; inverts H3. 
      exists (wfp_Par w p). splits... 
  (*par assoc*) inverts Ht...
    (*1*) destruct ap; inverts H3. 
      exists (wfp_Par (wfp_Par w q) r). splits...
    (*2*) inverts H2...
      destruct ap; inverts H3...
       exists (wfp_Par (wfp_Par p w) r). splits...
    (*2*) destruct aq0; inverts H3...
       exists (wfp_Par (wfp_Par p q) w). splits... 
  (*par assoc*) inverts Ht...
    (*1*) inverts H2... 
      destruct ap0; inverts H3. 
      exists (wfp_Par w (wfp_Par q r)). splits...
    (*2*) destruct aq; inverts H3...
       exists (wfp_Par p (wfp_Par w r)). splits...
    (*2*) destruct aq; inverts H3...
       exists (wfp_Par p (wfp_Par q w)). splits... 
  (*refl*) exists p'. splits...
Qed.

Lemma congr_tau : tau_relation wfcgr. 
Proof with (try solve [applys* t_step]; automate).
  introv Hr; inductions Hr. 

  Focus 2.
  introv Htp.
  lets (q' & Htq & Hcp'q') : IHHr1 Htp.  
  lets (r' & Htr & Hcq'r') : IHHr2 Htq.
  exists r'; splits~...
  etransitivity; eassumption.

  inductions H; introv Htp; try solve [inverts Htp].
 
  (* Par base *)
  rename IHwfcgr_step1 into HIp; 
  rename IHwfcgr_step2 into HIp'. 
  inverts Htp...
    (* Tau via Act on first *)
    lets (p'' & Heq) : wfp_agent_tau H4; subst; inverts H5.
    lets (q'' & Htq & Hcp''q'') : HIp H4; clear HIp.
    exists (wfp_Par q'' q'); splits...
      applys~ wfcgr_par... 
    lets (p'' & Heq) : wfp_agent_tau H4; subst; inverts H5.
    lets (q'' & Htq & Hcp''q'') : HIp' H4.
    exists (wfp_Par p' q''); splits...
      applys~ wfcgr_par... 

    (* Real Tau with output on first *)
    lets (q0' & q0'' & Htq0 & Hcq0' & Hcq0'') : congr_out p' H4...
    lets (fq1 & Htq1 & Heq) : congr_in_strong q' H5...
    exists (wfp_Par q0' (wfp_inst_Abs fq1 q0'')); splits...
      applys wfcgr_par...
        etransitivity. 
          apply Heq. applys~ wfcgr_inst_Abs. 

    (* Real Tau with output on second *)
    rename p into p0, q into p1, p' into q0, q' into q1,
           q'' into p1'', q'0 into p1', fp into fp0.
    apply (t_step _ wfcgr_step) in H.
    apply (t_step _ wfcgr_step) in H0.

    lets (q1' & q1'' & Htq1 & Hcq1' & Hcq1'') : congr_out H0 H5.
    lets (fq0 & Htq0 & Heq) : (congr_in_strong p0 q0 a fp0 H H4).
    exists (wfp_Par (wfp_inst_Abs fq0 q1'') q1'); splits; simpls.
      applys~ (TrTau2 a). 
      applys* wfcgr_par; try solve [inverts~ Hcq1']. 
        etransitivity.
          apply Heq. applys~ wfcgr_inst_Abs. 

  (* p || nil on the right *)
  exists (wfp_Par p' wfp_Nil); splits...

  (* p || nil on the left *)
  inverts Htp... 
    destruct ap; inverts H3...
      exists w; splits...
    inverts H2...
    inverts H3... 
    inverts H3...
  (* Commutativity of Par *)    
  inverts Htp...
    destruct ap; inverts H3...
      exists (wfp_Par q w); splits...
    destruct aq; inverts H3...
      exists (wfp_Par w p); splits...    
    exists (wfp_Par (wfp_inst_Abs fq p'') p'0); splits...
    exists (wfp_Par q' (wfp_inst_Abs fp q'')); splits...

  (* Associativity 1 - p || (q || r) and (p || q) || r *)
  inverts Htp...
    (* p -- Tau ->> via Act on first *)
    destruct ap; inverts H3.
    exists (wfp_Par (wfp_Par w q) r); splits...   

    (* q || r -- Tau ->> via Act on second *)
    destruct aq; inverts H3.
    inverts H2...
      (* q does a Tau via Act on first *)
      destruct ap; inverts H4.
      exists (wfp_Par (wfp_Par p w0) r); splits...
      (* r does a Tau via Act on second *)
      destruct aq; inverts H4.
      exists (wfp_Par (wfp_Par p q) w0); splits...
      (* Real Tau, output on q *)
      rename fq into fr; rename p' into q'; rename p'' into q''.
      apply TrAct2 with (p := p) in H3.      
      exists (wfp_Par (wfp_Par p q') (wfp_inst_Abs fr q'')); splits; simpls...
      (* Real Tau, output on r *)
      rename fp into fq; rename q' into r'; rename q'' into r''.
      apply TrAct2 with (p := p) in H3.      
      exists (wfp_Par (wfp_inst_Abs (AbsPar2 p fq) r'') r'); splits...      
        simpls... 

    (* Real Tau with output on p, input on q || r *)
    rename p'0 into p', fq into fqr.
      (* q does an input *)
      inverts H3...
      destruct ap; inverts H5... rename a0 into fq.
      assert ({{wfp_Par p q -- Tau ->> AP (wfp_Par p' (wfp_inst_Abs fq p''))}}) by applys~ (TrTau1 a).
      exists (wfp_Par (wfp_Par p' (wfp_inst_Abs fq p'')) r); splits; simpl...
      (* r does an input *)
      destruct aq; inverts H5... rename a0 into fr.
      apply act1_out with (p2 := q) in H2.
      exists (wfp_Par (wfp_Par p' q) (wfp_inst_Abs fr p'')); splits; simpl...

    (* Real Tau with input on p, output on q || r *)
    rename q'' into qr'', q' into qr'. inverts H3...
      (* q does an output *)
      destruct ap; inverts H5. 
      assert ({{wfp_Par p q -- Tau ->> AP (wfp_Par (wfp_inst_Abs fp qr'') w)}}) by applys~ (TrTau2 a).
      exists (wfp_Par (wfp_Par (wfp_inst_Abs fp qr'') w) r); splits...
      (* r does an output *)
      destruct aq; inverts H5. 
      apply act1_in with (p2 := q) in H2.
      exists (wfp_Par (wfp_inst_Abs (AbsPar1 fp q) qr'') w); splits... simpls...

  (* Associativity 2 - (p || q) || r and p || (q || r) *)
  inverts Htp...

    (* p || q -- Tau ->> via Act on first *)
    destruct ap; inverts H3.
    inverts H2...
      (* p does a Tau via Act on first *)
      destruct ap; inverts H4...
      exists (wfp_Par w0 (wfp_Par q r)); splits...
      (* q does a Tau via Act on second *)
      destruct aq; inverts H4...
      exists (wfp_Par p (wfp_Par w0 r)); splits...
      (* Real Tau, output on q *)
      apply act1_in with (p2 := r) in H4.      
      exists (wfp_Par p' (wfp_inst_Abs (AbsPar1 fq r) p'')); splits... simpls...
      (* Real Tau, output on q *)
      apply act1_out with (p2 := r) in H4.      
      exists (wfp_Par (wfp_inst_Abs fp q'') (wfp_Par q' r)); splits... 

    (* r -- Tau ->> via Act on second *)
    destruct aq; inverts H3.
    exists (wfp_Par p (wfp_Par q w)); splits...

    (* Real Tau with input on r, output on p || q *)
    rename p'' into pq'', p'0 into pq', fq into fr.
      (* p does an output *)
      inverts H2...
      destruct ap; inverts H5. 
      apply act2_in with (p1 := q) in H3.
      exists (wfp_Par w (wfp_inst_Abs (AbsPar2 q fr) pq'')); splits... simpls...
      (* q does an output *)
      destruct aq; inverts H5.
      assert ({{wfp_Par q r -- Tau ->> AP (wfp_Par w (wfp_inst_Abs fr pq''))}}) by applys~ (TrTau1 a).
      exists (wfp_Par p (wfp_Par w (wfp_inst_Abs fr pq''))); splits...

    (* Real Tau with output on r, input on p || q *)
    rename q'' into r'', q' into r', fp into fpq.
    inverts H2...
      (* p does an input *)
      destruct ap; inverts H5. rename a0 into fp.
      apply act2_out with (p1 := q) in H3.
      exists (wfp_Par (wfp_inst_Abs fp r'') (wfp_Par q r')); splits... simpl...
      (* q does an input *)
      destruct aq; inverts H5. rename a0 into fq.
      assert ({{wfp_Par q r -- Tau ->> AP (wfp_Par (wfp_inst_Abs fq r'') r')}})...
      exists (wfp_Par p (wfp_Par (wfp_inst_Abs fq r'') r')). splits... simpls...

  exists p'; splits... 
Qed.

Lemma congr_out_normal_strong : forall a (p p' p'' q: wfprocess), 
      p ≡* q -> {{p -- LabOut a p'' ->> AP p'}} ->
      exists (q' q'' : wfprocess), {{q -- LabOut a q'' ->> AP q'}} /\
      (∀ (m : chan) (X : var), (wfp_Par (wfp_Abs m X p'') p') ≡* (wfp_Par (wfp_Abs m X q'') q')).                    
Proof with (try solve [applys* t_step]; try apply wfcgr_par; automate).
  introv Hr; gen p' p''; inductions Hr.

  Focus 2.
  introv Htx.
  lets (q' & q'' & Hty & Heqy) : IHHr1 Htx.
  lets (r' & r'' & Htr & Heqr) : IHHr2 Hty.
  exists r' r''; splits...
  introv; etransitivity; jauto.

  inductions H.

  (* Send *)
  introv Ht; inverts Ht...
  exists wfp_Nil q; splits...
    introv... 

  (* Receive *)
  introv Ht; inverts Ht.

  (* Par base *)
  introv Ht; inverts Ht...
    destruct ap; inverts H5...
    lets (q0' & q0'' & Htq & Hc' & Hc'') : congr_out p' H4...
    exists (wfp_Par q0' q') q0''; splits...
      introv; simpls... 
        apply wfcgr_abs...
        apply wfcgr_par... 
    destruct aq; inverts H5.
    lets (q1' & q1'' & Htq & Hc' & Hc'') : congr_out q' H4...
    exists (wfp_Par p' q1') q1''; splits...
      introv; simpls...
        apply wfcgr_abs...
        apply wfcgr_par...

  (* p || nil on the right *)  
  introv Ht.
  exists (wfp_Par p' wfp_Nil) p''; splits; introv...

  (* p || nil on the left *)  
  introv Ht; inverts Ht...
    destruct ap; inverts H3.
    exists w p''; splits; introv...
    inverts H2.

  (* Par com *)
  introv Ht; inverts Ht...
    destruct ap; inverts H3.
    exists (wfp_Par q w) p''; splits; introv...
    destruct aq; inverts H3.
    exists (wfp_Par w p) p''; splits; introv...
 
  (* Associativity 1 - p || (q || r) and (p || q) || r *)
  introv Htp; inverts Htp...
    (* Output on p *)
    destruct ap; inverts H3.
    exists (wfp_Par (wfp_Par w q) r) p''; splits; introv...
    inverts H2...
      (* Output on q *)
      destruct ap; inverts H3.  
      exists (wfp_Par (wfp_Par p w) r) p''; splits; introv...
      (* Output on r *)
      destruct aq0; inverts H3.
      exists (wfp_Par (wfp_Par p q) w) p''; splits; introv...

  (* Associativity 2 - (p || q) || r and p || (q || r) *)
  introv Htp; inverts Htp...
    inverts H2...
    (* Output on p *)
    destruct ap0; inverts H3.
    exists (wfp_Par w (wfp_Par q r)) p''; splits; introv...
    (* Output on q *)
    destruct aq; inverts H3.
    exists (wfp_Par p (wfp_Par w r)) p''; splits; introv...
    (* Output on r *)
    destruct aq; inverts H3.
    exists (wfp_Par p (wfp_Par q w)) p''; splits; introv...

  introv Ht; exists p' p''; splits; introv...
Qed.

Lemma congr_out_normal : out_normal_relation wfcgr.
Proof.
  introv Hr Ht.
  lets HDexM : congr_out_normal_strong Hr Ht.
  destruct HDexM as (q' & q'' & Htq & Heq).
  exists q' q''; splits~; introv _ _ _ _; jauto.
Qed.

Lemma congr_in_normal : in_normal_relation wfcgr.
Proof.
  introv Hr Ht. 
  lets Hin : congr_in_strong Hr Ht.
  destruct Hin as (fq & Htq & Hyp).
  exists* fq.
Qed.

Lemma congr_closed : closed_relation wfcgr. 
Proof.
  introv Hr Ht. 
  lets Hin : congr_in_strong Hr Ht. 
  destruct Hin as (fq & Htq & Hyp).
  exists* fq. 
Qed.

Lemma congr_out_context_strong : 
  forall p q : wfprocess, p ≡* q -> 
    forall a (p' p'' : wfprocess), 
      {{p--LabOut a p''->>(AP p')}} ->
        exists (q' q'' : wfprocess), 
          ({{q--LabOut a q''->>(AP q')}} /\ forall X s,
            ((wfp_Par (wfp_subst p'' X s) p') ≡* (wfp_Par (wfp_subst q'' X s) q'))). 
Proof with (try apply wfcgr_subst1; try solve [applys* t_step]; automate).
  introv Hr; inductions Hr.
  
  Focus 2.
  introv Htx.
  specializes IHHr1 Htx. destruct IHHr1 as (q' & q'' & Htq & Hypq).
  specializes IHHr2 Htq. destruct IHHr2 as (r' & r'' & Htr & Hypr).
  exists r' r''; splits...
  introv; specialize (Hypq X s); specialize (Hypr X s).
  etransitivity; eassumption.
  
  inductions H; introv Ht.

    (* send *)
    rename IHwfcgr_step into IH.
    inverts Ht...
    exists wfp_Nil q; splits... 
      introv; apply wfcgr_par... 

    (* abs *)
    inverts Ht.
 
    (* par *)
    rename IHwfcgr_step1 into IH; rename IHwfcgr_step2 into IH1.
    inverts Ht...
      destruct ap; inverts H5.
      lets (q1 & q2 & Htq & Hypq) : IH H4; clear IH.
      exists (wfp_Par q1 q') q2; splits; introv...
        introv; specialize (Hypq X s).
        etransitivity.
          constructor. apply WfCgrPar_assoc1.      
          etransitivity.
            Focus 2. constructor. apply WfCgrPar_assoc2.
            applys wfcgr_par... 
      destruct aq; inverts H5.
      lets (q1' & q1'' & Htq1 & Hypq1) : IH1 H4; clear IH1.
      exists (wfp_Par p' q1') q1''; splits...
        introv; specialize (Hypq1 X s).
        assert (Hcp : (wfp_Par (wfp_subst p'' X s) (wfp_Par w p)) ≡* (wfp_Par (wfp_subst p'' X s) (wfp_Par p w)))...
        assert (Hcq : (wfp_Par (wfp_subst q1'' X s) (wfp_Par q1' p')) ≡* (wfp_Par (wfp_subst q1'' X s) (wfp_Par p' q1')))...
        etransitivity. symmetry. apply Hcp.
        etransitivity. Focus 2. apply Hcq.        
         etransitivity.
          constructor. apply WfCgrPar_assoc1.      
          etransitivity.
            Focus 2. constructor. apply WfCgrPar_assoc2.
            applys~ wfcgr_par... 

    (* nil1 *)
    exists (wfp_Par p' wfp_Nil) p''; splits; introv...

    (* nil2 *)
    inverts Ht...
      destruct ap; inverts H3...
      exists w p''; splits; introv...
      inverts H2.

   (* comm *)
   inverts Ht... 
     destruct ap; inverts H3.
     exists (wfp_Par q w) p''; splits; introv...
     destruct aq; inverts H3.
     exists (wfp_Par w p) p''; splits; introv...

   (* assoc1 *)
   inverts Ht...
     (* Output on p *)
     destruct ap; inverts H3.
     exists (wfp_Par (wfp_Par w q) r) p''; splits; introv...
     inverts H2...
       (* Output on q *)
       destruct ap; inverts H3.
       exists (wfp_Par (wfp_Par p w) r) p''; splits; introv...
       (* Output on r *)
       destruct aq0; inverts H3.
       exists (wfp_Par (wfp_Par p q) w) p''; splits; introv...

   (* assoc2 *)
   inverts Ht...
     inverts H2...
     (* Output on p *)
     destruct ap0; inverts H3.
     exists (wfp_Par w (wfp_Par q r)) p''; splits; introv...
     (* Output on q *)
     destruct aq; inverts H3.
     exists (wfp_Par p (wfp_Par w r)) p''; splits; introv...
     (* Output on r *)
     destruct aq; inverts H3.
     exists (wfp_Par p (wfp_Par q w)) p''; splits~; introv...

   (* refl *)
   exists p' p''; splits; introv...
Qed.

Lemma congr_out_context : out_context_relation wfcgr.
Proof.
  introv Hc Htp.
  lets (q' & q'' & Htq & Hyp) : congr_out_context_strong Hc Htp.
  exists q' q''; splits~.
Qed.

Hint Resolve congr_in congr_out congr_rem congr_tau
             congr_in_normal congr_out_normal congr_closed 
             congr_out_context : congr_bisim.

Lemma congr_IObisimulation : IO_bisimulation wfcgr. 
Proof.
  splits; eauto with congr_bisim.
  apply wfcgr_is_eq_rel.
Qed.

Lemma congr_HObisimulation: HigherOrder_bisimulation wfcgr.
Proof.
  splits; auto with congr_bisim.
  apply wfcgr_is_eq_rel.
Qed.

Lemma congr_NORbisimulation: Normal_bisimulation wfcgr.
Proof.
  splits; auto with congr_bisim.
  apply wfcgr_is_eq_rel.
Qed.
   
Lemma congr_ONORbisimulation : OpenNormal_bisimulation wfcgr.
Proof.
  splits; auto with congr_bisim.
  apply wfcgr_is_eq_rel.
Qed.

Lemma congr_CONbisimulation : Context_bisimulation wfcgr.
Proof.
  splits; auto with congr_bisim.
  apply wfcgr_is_eq_rel.
Qed.

Hint Resolve congr_IObisimulation congr_HObisimulation
             congr_NORbisimulation congr_ONORbisimulation 
             congr_CONbisimulation : congr_bisim.
